from typing import List, Tuple

import torch as to
from torch.utils.data import Dataset


class GraphDataset(Dataset):
    def __init__(self, raw_dataset: List[Tuple[to.Tensor, to.Tensor]]) -> None:
        self.features = self._to_list(raw_dataset)
        self.labels = self._preprocess_labels(raw_dataset)

    def __len__(self) -> int:
        return len(self.features)

    def __getitem__(self, index: int) -> Tuple:
        return self.features[index], self.labels[index]

    @staticmethod
    def _to_list(raw_dataset: List[Tuple[to.Tensor, to.Tensor]]) -> List[Tuple[to.Tensor, to.Tensor]]:
        return [(raw_dataset[index][0], raw_dataset[index][1].view(-1)) for index in range(len(raw_dataset))]

    def _preprocess_labels(self, raw_dataset: List[Tuple[to.Tensor, to.Tensor]]) -> List[to.Tensor]:
        labels = self._extract_labels(raw_dataset)
        labels_preprocessed = []
        for index in range(len(labels)):
            upper_triangular_indices = to.triu(to.ones_like(labels[index]) - to.eye(labels[index].size()[0],
                                                                                    labels[index].size()[1])) == 1
            labels_preprocessed.append(labels[index][upper_triangular_indices].float())
        return labels_preprocessed

    @staticmethod
    def _extract_labels(raw_dataset: List[Tuple[to.Tensor, to.Tensor]]) -> List[to.Tensor]:
        return [raw_dataset[index][1] for index in range(len(raw_dataset))]



from unittest import TestCase

import numpy as np

from message_passing_nn.model.edge import Edge
from message_passing_nn.model.node import Node
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestEdge(TestCase):
    def setUp(self) -> None:
        node_features = BASE_GRAPH_NODE_FEATURES
        labels = BASE_GRAPH.view(-1)
        start_node = Node(node_features, labels, 2)
        end_node = Node(node_features, labels, 0)
        self.edge = Edge(start_node, end_node)

    def test_get_edge_slice(self):
        # Given
        edge_slice_expected = (2, 0)

        # When
        edge_slice = self.edge.get_edge_slice()

        # Then
        self.assertEqual(edge_slice_expected, edge_slice)

    def test_get_start_node_neighbors_without_end_node(self):
        # Given
        start_node_neighbors_expected = np.array([1, 3])

        # When
        start_node_neighbors = self.edge.get_start_node_neighbors_without_end_node()[0]

        # Then
        self.assertTrue(np.array_equal(start_node_neighbors_expected, start_node_neighbors))



from unittest import TestCase

import torch as to

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.graph import Graph
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestGraphPreprocessor(TestCase):
    def setUp(self) -> None:
        self.data_preprocessor = DataPreprocessor()

    def test_train_validation_test_split(self):
        # Given
        dataset_length = 10
        features = BASE_GRAPH_NODE_FEATURES
        labels = BASE_GRAPH
        raw_dataset = [(features, labels) for i in range(dataset_length)]
        train_validation_test_split_expected = [7, 2, 1]

        # When
        train_validation_test_split = self.data_preprocessor.train_validation_test_split(raw_dataset,
                                                                                         batch_size=1,
                                                                                         validation_split=0.2,
                                                                                         test_split=0.1)
        train_validation_test_split = [len(dataset) for dataset in train_validation_test_split]

        # Then
        self.assertEqual(train_validation_test_split_expected, train_validation_test_split)

    def test_extract_initialization_graph(self):
        # Given
        dataset_length = 1
        features = BASE_GRAPH_NODE_FEATURES
        labels = BASE_GRAPH
        raw_dataset = [(features, labels) for i in range(dataset_length)]
        initialization_graph_expected = Graph(labels, features)

        # When
        initialization_graph = self.data_preprocessor.extract_initialization_graph(raw_dataset)

        # Then
        self.assertEqual(initialization_graph_expected, initialization_graph)

    def test_flatten_when_sizes_match(self):
        # Given
        dataset_length = 2
        labels = BASE_GRAPH.view(-1)
        tensors = to.cat((labels, labels))
        tensors_flattened_expected = tensors.view(-1)

        # When
        tensors_flattened = self.data_preprocessor.flatten(tensors, desired_size=dataset_length*len(labels))

        # Then
        self.assertTrue(to.allclose(tensors_flattened_expected, tensors_flattened))

    def test_flatten_when_sizes_do_not_match(self):
        # Given
        dataset_length = 3
        labels = BASE_GRAPH.view(-1)
        tensors = to.cat((labels, labels))
        tensors_flattened_expected = to.cat((tensors.view(-1), to.zeros_like(labels)))

        # When
        tensors_flattened = self.data_preprocessor.flatten(tensors, desired_size=dataset_length*len(labels))

        # Then
        self.assertTrue(to.allclose(tensors_flattened_expected, tensors_flattened))



from unittest import TestCase

import torch as to

from message_passing_nn.model.message_gru import MessageGRU


class TestMessage(TestCase):
    def setUp(self) -> None:
        device = "cpu"
        self.message = MessageGRU(device)
        self.message.update_gate = to.tensor([1, 2])
        self.message.previous_messages = to.tensor([-2, 2])
        self.message.current_memory = to.tensor([-1, 1])

    def test_compose(self):
        # Given
        message_value_expected = to.tensor([-1, 0]).float()

        # When
        self.message.compose()

        # Then
        self.assertTrue(to.allclose(message_value_expected, self.message.value))



from torch import nn

loss_functions = {
    "MSE": nn.MSELoss(),
    "L1": nn.L1Loss(),
    "CrossEntropy": nn.CrossEntropyLoss(),
    "CTC": nn.CTCLoss(),
    "NLL": nn.NLLLoss(),
    "PoissonNLL": nn.PoissonNLLLoss(),
    "KLDiv": nn.KLDivLoss(),
    "BCE": nn.BCELoss(),
    "BCEWithLogits": nn.BCEWithLogitsLoss(),
    "MarginRanking": nn.MarginRankingLoss(),
    "HingeEmbedding": nn.HingeEmbeddingLoss(),
    "MultiLabelMargin": nn.MultiLabelMarginLoss(),
    "SmoothL1": nn.SmoothL1Loss(),
    "SoftMargin": nn.SoftMarginLoss(),
    "MultiLabelSoftMargin": nn.MultiLabelSoftMarginLoss(),
    "CosineEmbedding": nn.CosineEmbeddingLoss(),
    "MultiMargin": nn.MultiMarginLoss(),
    "TripletMargin": nn.TripletMarginLoss()
}



import logging

from torch.nn.modules.module import Module

from message_passing_nn.utils.loss_functions import loss_functions


class LossFunctionSelector:
    def __init__(self) -> None:
        pass

    @staticmethod
    def load_loss_function(loss_function_selection: str) -> Module:
        if loss_function_selection in loss_functions:
            return loss_functions[loss_function_selection]
        else:
            get_logger().info("The " + loss_function_selection + " is not available")
            raise Exception


def get_logger() -> logging.Logger:
    return logging.getLogger('message_passing_nn')



import numpy as np
import torch as to


class Node:
    def __init__(self, all_node_features: to.Tensor, adjacency_matrix: to.Tensor, node_id: to.Tensor):
        self.node_id = self._get_integer(node_id)
        self.features = self._get_features_of_specific_node(all_node_features)
        self.neighbors = self._get_neighbors(adjacency_matrix)
        self.neighbors_count = len(self.neighbors)

    def _get_features_of_specific_node(self, all_node_features: to.Tensor) -> to.Tensor:
        return all_node_features[self.node_id]

    def _get_neighbors(self, adjacency_matrix: to.Tensor) -> to.Tensor:
        number_of_nodes = self._get_integer(np.sqrt(adjacency_matrix.size()[-1]))
        adjacency_matrix = adjacency_matrix.view(number_of_nodes, number_of_nodes)
        return to.nonzero(adjacency_matrix[self.node_id], as_tuple=True)[0]

    @staticmethod
    def _get_integer(field: to.Tensor) -> int:
        return int(field)




from unittest import TestCase

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.graph import Graph
from message_passing_nn.model.graph_encoder import GraphEncoder
from message_passing_nn.trainer.model_trainer import ModelTrainer
from tests.fixtures.matrices_and_vectors import BASE_GRAPH, BASE_GRAPH_NODE_FEATURES


class TestModelTrainer(TestCase):
    def setUp(self) -> None:
        time_steps = 1
        loss_function = "MSE"
        optimizer = "SGD"
        device = "cpu"
        self.configuration_dictionary = {"loss_function": loss_function,
                                         "optimizer": optimizer,
                                         "time_steps": time_steps}
        self.model_trainer = ModelTrainer(GraphEncoder, device)

    def test_instantiate_attributes(self):
        # Given
        initialization_graph = Graph(BASE_GRAPH, BASE_GRAPH_NODE_FEATURES)

        # When
        self.model_trainer.instantiate_attributes(initialization_graph, self.configuration_dictionary)

        # Then
        self.assertTrue(self.model_trainer.model.number_of_nodes == initialization_graph.number_of_nodes)
        self.assertTrue(
            self.model_trainer.model.number_of_node_features == initialization_graph.number_of_node_features)
        self.assertTrue(self.model_trainer.optimizer.param_groups)

    def test_do_train(self):
        # Given
        self.model_trainer.instantiate_attributes(Graph(BASE_GRAPH, BASE_GRAPH_NODE_FEATURES),
                                                  self.configuration_dictionary)
        raw_dataset = [(BASE_GRAPH_NODE_FEATURES, BASE_GRAPH)]
        training_data, _, _ = DataPreprocessor.train_validation_test_split(raw_dataset, 1, 0.0, 0.0)

        # When
        training_loss = self.model_trainer.do_train(training_data=training_data, epoch=1)

        # Then
        self.assertTrue(training_loss > 0.0)

    def test_do_evaluate(self):
        # Given
        self.model_trainer.instantiate_attributes(Graph(BASE_GRAPH, BASE_GRAPH_NODE_FEATURES),
                                                  self.configuration_dictionary)
        raw_dataset = [(BASE_GRAPH_NODE_FEATURES, BASE_GRAPH)]
        training_data, _, _ = DataPreprocessor.train_validation_test_split(raw_dataset, 1, 0.0, 0.0)

        # When
        validation_loss = self.model_trainer.do_evaluate(evaluation_data=training_data, epoch=1)

        # Then
        self.assertTrue(validation_loss > 0.0)



import logging
from typing import Tuple, List

import torch as to
from torch.utils.data import DataLoader

from message_passing_nn.data.graph_dataset import GraphDataset
from message_passing_nn.data.preprocessor import Preprocessor
from message_passing_nn.model.graph import Graph


class DataPreprocessor(Preprocessor):
    def __init__(self):
        super().__init__()

    @staticmethod
    def train_validation_test_split(raw_dataset: List[Tuple[to.Tensor, to.Tensor]],
                                    batch_size: int,
                                    validation_split: float = 0.2,
                                    test_split: float = 0.1) -> Tuple[DataLoader, DataLoader, DataLoader]:
        test_index, validation_index = DataPreprocessor._get_validation_and_test_indexes(raw_dataset, test_split,
                                                                                         validation_split)
        training_data = DataLoader(GraphDataset(raw_dataset[:validation_index]), batch_size)
        if validation_split:
            validation_data = DataLoader(GraphDataset(raw_dataset[validation_index:test_index]), batch_size)
        else:
            validation_data = DataLoader(GraphDataset([]))
        if test_split:
            test_data = DataLoader(GraphDataset(raw_dataset[test_index:]), batch_size)
        else:
            test_data = DataLoader(GraphDataset([]))
        return training_data, validation_data, test_data

    @staticmethod
    def extract_initialization_graph(raw_dataset: List[Tuple[to.Tensor, to.Tensor]]) -> Graph:
        return Graph(raw_dataset[0][1], raw_dataset[0][0])

    @staticmethod
    def flatten(tensors: to.Tensor, desired_size: int = 0) -> to.Tensor:
        flattened_tensor = tensors.view(-1)
        if 0 < desired_size != len(flattened_tensor):
            flattened_tensor = DataPreprocessor._pad_zeros(flattened_tensor, desired_size)
        return flattened_tensor

    @staticmethod
    def _pad_zeros(flattened_tensor: to.Tensor, desired_size: int) -> to.Tensor:
        size_difference = abs(len(flattened_tensor) - desired_size)
        flattened_tensor = to.cat((flattened_tensor, to.zeros(size_difference)))
        return flattened_tensor

    @staticmethod
    def _get_validation_and_test_indexes(raw_dataset: List[Tuple[to.Tensor, to.Tensor]],
                                         test_split: float,
                                         validation_split: float) -> Tuple[int, int]:
        validation_index = int((1 - validation_split - test_split) * len(raw_dataset))
        test_index = int((1 - test_split) * len(raw_dataset))
        return test_index, validation_index

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



from typing import Tuple, Any

import torch as to

from message_passing_nn.model.node import Node


class Edge:
    def __init__(self, start_node: Node, end_node: Node) -> None:
        self.start_node = start_node
        self.end_node = end_node

    def get_edge_slice(self) -> Tuple:
        return self.start_node.node_id, self._get_integer(self.end_node.node_id)

    def get_start_node_neighbors_without_end_node(self) -> Tuple:
        return self._remove_end_node_from_start_node_neighbors(), [self.start_node.node_id]

    def _remove_end_node_from_start_node_neighbors(self) -> to.tensor:
        end_node_index = (self.start_node.neighbors == self.end_node.node_id).nonzero()[0][0].item()
        return to.cat((self.start_node.neighbors[:end_node_index],
                       self.start_node.neighbors[end_node_index + 1:])).tolist()

    @staticmethod
    def _get_integer(field: Any) -> int:
        return int(field)



from torch import nn

loss_functions = {
    "MSE": nn.MSELoss,
    "L1": nn.L1Loss,
    "CrossEntropy": nn.CrossEntropyLoss,
    "CTC": nn.CTCLoss,
    "NLL": nn.NLLLoss,
    "PoissonNLL": nn.PoissonNLLLoss,
    "KLDiv": nn.KLDivLoss,
    "BCE": nn.BCELoss,
    "BCEWithLogits": nn.BCEWithLogitsLoss,
    "MarginRanking": nn.MarginRankingLoss,
    "HingeEmbedding": nn.HingeEmbeddingLoss,
    "MultiLabelMargin": nn.MultiLabelMarginLoss,
    "SmoothL1": nn.SmoothL1Loss,
    "SoftMargin": nn.SoftMarginLoss,
    "MultiLabelSoftMargin": nn.MultiLabelSoftMarginLoss,
    "CosineEmbedding": nn.CosineEmbeddingLoss,
    "MultiMargin": nn.MultiMarginLoss,
    "TripletMargin": nn.TripletMarginLoss
}



from abc import ABCMeta, abstractmethod
from typing import Any, Tuple, List

from torch.utils.data import DataLoader

from message_passing_nn.model.graph import Graph


class Preprocessor(metaclass=ABCMeta):
    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def train_validation_test_split(raw_dataset: Any,
                                    batch_size: int,
                                    validation_split: float = 0.2,
                                    test_split: float = 0.1) -> Tuple[DataLoader, DataLoader, DataLoader]:
        pass

    @staticmethod
    @abstractmethod
    def extract_initialization_graph(raw_dataset: Any) -> Graph:
        pass

    @staticmethod
    @abstractmethod
    def flatten(tensors: List[Any], desired_size: Any = 0) -> Any:
        pass



import logging

from torch.optim.optimizer import Optimizer

from message_passing_nn.utils.optimizers import optimizers


class OptimizerSelector:
    def __init__(self) -> None:
        pass

    @staticmethod
    def load_optimizer(optimizer_selection: str) -> Optimizer:
        if optimizer_selection in optimizers:
            return optimizers[optimizer_selection]
        else:
            get_logger().info("The " + optimizer_selection + " is not available")
            raise Exception


def get_logger() -> logging.Logger:
    return logging.getLogger('message_passing_nn')



import logging
from typing import List, Dict

import numpy as np

from message_passing_nn.fixtures.characters import GRID_SEARCH_SEPARATION_CHARACTER


class GridSearchParametersParser:
    def __init__(self) -> None:
        pass

    def get_grid_search_dictionary(self,
                                   epochs: str,
                                   loss_function_selection: str,
                                   optimizer_selection: str,
                                   batch_size: str,
                                   validation_split: str,
                                   test_split: str,
                                   time_steps: str,
                                   validation_period: str) -> Dict:
        return {
            'epochs': self._parse_integer_range(epochs),
            'loss_function': self._parse_string_selections(loss_function_selection),
            'optimizer': self._parse_string_selections(optimizer_selection),
            'batch_size': self._parse_integer_range(batch_size),
            'validation_split': self._parse_float_range(validation_split),
            'test_split': self._parse_float_range(test_split),
            'time_steps': self._parse_integer_range(time_steps),
            'validation_period': self._parse_integer_range(validation_period)
        }

    def _parse_integer_range(self, field: str) -> List[int]:
        integer_range = field.split(GRID_SEARCH_SEPARATION_CHARACTER)
        if len(integer_range) == 1:
            return [int(integer_range[0])]
        elif len(integer_range) == 3:
            min_range, max_range, number_of_values = integer_range
            integer_range = np.linspace(int(min_range), int(max_range), int(number_of_values))
            return [int(number) for number in integer_range]
        else:
            self.get_logger().info("Incorrect values for integer range. Please either provide "
                                   "a single integer or three integers separated by & (min&max&values)")
            raise Exception

    def _parse_float_range(self, field: str) -> List[float]:
        float_range = field.split(GRID_SEARCH_SEPARATION_CHARACTER)
        if len(float_range) == 1:
            return [float(float_range[0])]
        elif len(float_range) == 3:
            min_range, max_range, number_of_values = float_range
            float_range = np.linspace(float(min_range), float(max_range), int(number_of_values))
            return [float(number) for number in float_range]
        else:
            self.get_logger().info("Incorrect values for float range. Please either provide "
                                   "a single float or two floats and an integer separated by & (min&max&values)")
            raise Exception

    @staticmethod
    def _parse_string_selections(loss_function_selection: str) -> List[str]:
        return loss_function_selection.split(GRID_SEARCH_SEPARATION_CHARACTER)

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



import torch as to
import torch.nn as nn

from message_passing_nn.data.data_preprocessor import DataPreprocessor
from message_passing_nn.model.edge import Edge
from message_passing_nn.model.graph import Graph
from message_passing_nn.model.message_gru import MessageGRU
from message_passing_nn.model.node import Node


class GraphEncoder(nn.Module):
    def __init__(self,
                 time_steps: int,
                 number_of_nodes: int,
                 number_of_node_features: int,
                 fully_connected_layer_input_size: int,
                 fully_connected_layer_output_size: int,
                 device: str) -> None:
        super(GraphEncoder, self).__init__()
        self.time_steps = time_steps
        self.number_of_nodes = number_of_nodes
        self.number_of_node_features = number_of_node_features
        self.fully_connected_layer_input_size = fully_connected_layer_input_size
        self.fully_connected_layer_output_size = fully_connected_layer_output_size
        self.w_gru_update_gate_features = None
        self.w_gru_forget_gate_features = None
        self.w_gru_current_memory_message_features = None
        self.u_gru_update_gate = None
        self.u_gru_forget_gate = None
        self.u_gru_current_memory_message = None
        self.b_gru_update_gate = None
        self.b_gru_forget_gate = None
        self.b_gru_current_memory_message = None
        self.u_graph_node_features = None
        self.u_graph_neighbor_messages = None
        self.device = device
        self.linear = to.nn.Linear(self.fully_connected_layer_input_size, self.fully_connected_layer_output_size)
        self.sigmoid = to.nn.Sigmoid()

    @classmethod
    def of(cls,
           time_steps: int,
           number_of_nodes: int,
           number_of_node_features: int,
           fully_connected_layer_input_size: int,
           fully_connected_layer_output_size: int,
           device: str):
        return cls(time_steps,
                   number_of_nodes,
                   number_of_node_features,
                   fully_connected_layer_input_size,
                   fully_connected_layer_output_size,
                   device)

    def forward(self, node_features: to.Tensor, adjacency_matrix: to.Tensor, batch_size: int) -> to.Tensor:
        outputs = to.zeros(batch_size, self.fully_connected_layer_output_size).to(self.device)
        for batch in range(batch_size):
            outputs[batch] = self.sigmoid(
                self.linear(
                    DataPreprocessor.flatten(
                        self.encode(node_features[batch], adjacency_matrix[batch]))))
        return outputs

    def encode(self, node_features: to.Tensor, adjacency_matrix: to.Tensor) -> to.Tensor:
        messages = self._send_messages(node_features, adjacency_matrix)
        encodings = self._encode_nodes(node_features, messages)
        return encodings

    def initialize_tensors(self, graph: Graph) -> None:
        base_4d_tensor_shape = [graph.number_of_nodes,
                                graph.number_of_nodes,
                                graph.number_of_node_features,
                                graph.number_of_node_features]
        base_3d_tensor_shape = [graph.number_of_nodes,
                                graph.number_of_node_features,
                                graph.number_of_node_features]
        base_2d_tensor_shape = [graph.number_of_node_features]
        self.w_gru_update_gate_features = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.w_gru_forget_gate_features = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.w_gru_current_memory_message_features = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.u_gru_update_gate = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.u_gru_forget_gate = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.u_gru_current_memory_message = nn.Parameter(to.randn(base_4d_tensor_shape), requires_grad=True).to(self.device)
        self.b_gru_update_gate = nn.Parameter(to.randn(base_2d_tensor_shape), requires_grad=True).to(self.device)
        self.b_gru_forget_gate = nn.Parameter(to.randn(base_2d_tensor_shape), requires_grad=True).to(self.device)
        self.b_gru_current_memory_message = nn.Parameter(to.randn(base_2d_tensor_shape), requires_grad=True).to(self.device)
        self.u_graph_node_features = nn.Parameter(to.randn(base_3d_tensor_shape), requires_grad=True).to(self.device)
        self.u_graph_neighbor_messages = nn.Parameter(to.randn(base_3d_tensor_shape), requires_grad=True).to(self.device)

    def _send_messages(self, node_features: to.Tensor, adjacency_matrix: to.Tensor) -> to.Tensor:
        messages = to.zeros((self.number_of_nodes, self.number_of_nodes, self.number_of_node_features)).to(self.device)
        for step in range(self.time_steps):
            messages = self._compose_messages_from_nodes_to_targets(node_features,
                                                                    adjacency_matrix,
                                                                    messages)
        return messages

    def _encode_nodes(self, node_features: to.Tensor, messages: to.Tensor) -> to.Tensor:
        encoded_node = to.zeros(self.number_of_nodes, self.number_of_node_features).to(self.device)
        for node_id in range(self.number_of_nodes):
            encoded_node[node_id] += self._apply_recurrent_layer_for_node(node_features,
                                                                          messages,
                                                                          node_id)
        return encoded_node

    def _apply_recurrent_layer_for_node(self, node_features: to.Tensor, messages: to.Tensor, node_id: int) -> to.Tensor:
        node_encoding_features = self.u_graph_node_features[node_id].matmul(node_features[node_id])
        node_encoding_messages = self.u_graph_neighbor_messages[node_id].matmul(to.sum(messages[node_id], dim=0))
        return to.relu(node_encoding_features + node_encoding_messages)

    def _compose_messages_from_nodes_to_targets(self,
                                                node_features: to.Tensor,
                                                adjacency_matrix: to.Tensor,
                                                messages: to.Tensor) -> to.Tensor:
        new_messages = to.zeros(messages.shape).to(self.device)
        for node_id in range(self.number_of_nodes):
            node = self._create_node(node_features, adjacency_matrix, node_id)
            for end_node_id in node.neighbors:
                end_node = self._create_node(node_features, adjacency_matrix, end_node_id)
                edge = self._create_edge(node, end_node)
                edge_slice = edge.get_edge_slice()
                message = self._get_message_inputs(messages, node, edge, node_features, adjacency_matrix)
                message.compose()
                new_messages[edge_slice] = message.value
        return new_messages

    def _get_message_inputs(self,
                            messages: to.Tensor,
                            node: Node,
                            edge: Edge,
                            node_features: to.Tensor,
                            adjacency_matrix: to.Tensor) -> MessageGRU:
        message = self._create_message(self.device)
        message.previous_messages = self._get_messages_from_all_node_neighbors_except_target_summed(messages,
                                                                                                    node,
                                                                                                    edge)
        message.update_gate = self._pass_through_update_gate(messages, node, edge, node_features)
        message.current_memory = self._get_current_memory_message(messages, node, edge, node_features, adjacency_matrix)
        return message

    def _get_messages_from_all_node_neighbors_except_target_summed(self,
                                                                   messages: to.Tensor,
                                                                   node: Node,
                                                                   edge: Edge) -> to.Tensor:
        messages_from_the_other_neighbors = to.zeros(node.features.shape[0]).to(self.device)
        if node.neighbors_count > 1:
            neighbors_slice = edge.get_start_node_neighbors_without_end_node()
            messages_from_the_other_neighbors = to.sum(messages[neighbors_slice], dim=0)
        return messages_from_the_other_neighbors

    def _pass_through_update_gate(self,
                                  messages: to.Tensor,
                                  node: Node,
                                  edge: Edge,
                                  node_features: to.Tensor) -> to.Tensor:
        message_from_a_neighbor_other_than_target = self._get_messages_from_all_node_neighbors_except_target_summed(
            messages,
            node,
            edge)
        edge_slice = edge.get_edge_slice()
        update_gate_output = to.sigmoid(
            to.add(
                to.add(self.w_gru_update_gate_features[edge_slice].matmul(node_features[node.node_id]),
                       self.u_gru_update_gate[edge_slice].matmul(message_from_a_neighbor_other_than_target)),
                self.b_gru_update_gate))
        return update_gate_output

    def _get_current_memory_message(self,
                                    messages: to.Tensor,
                                    node: Node,
                                    edge: Edge,
                                    node_features: to.Tensor,
                                    adjacency_matrix: to.Tensor) -> to.Tensor:
        messages_passed_through_reset_gate = self._keep_or_reset_messages(messages,
                                                                          node,
                                                                          edge,
                                                                          node_features,
                                                                          adjacency_matrix)
        edge_slice = edge.get_edge_slice()
        current_memory_message = to.add(
            to.add(self.w_gru_current_memory_message_features[edge_slice].matmul(node_features[node.node_id]),
                   self.u_gru_current_memory_message[edge_slice].matmul(messages_passed_through_reset_gate)),
            self.b_gru_current_memory_message)
        return to.tanh(current_memory_message)

    def _keep_or_reset_messages(self,
                                messages: to.Tensor,
                                node: Node,
                                edge: Edge,
                                node_features: to.Tensor,
                                adjacency_matrix: to.Tensor) -> to.Tensor:
        edge_slice = edge.get_edge_slice()
        neighbors_slice = edge.get_start_node_neighbors_without_end_node()[0]
        messages_from_the_other_neighbors = to.zeros(node.features.shape[0]).to(self.device)
        for reset_node_index in neighbors_slice:
            reset_node = self._create_node(node_features, adjacency_matrix, reset_node_index)
            reset_edge = self._create_edge(node, reset_node)
            reset_edge_slice = reset_edge.get_edge_slice()
            reset_gate_output = self._pass_through_reset_gate(messages, node, reset_edge, node_features)
            messages_from_the_other_neighbors += to.mul(reset_gate_output, messages[reset_edge_slice])
        return self.u_gru_current_memory_message[edge_slice].matmul(messages_from_the_other_neighbors)

    def _pass_through_reset_gate(self,
                                 messages: to.Tensor,
                                 node: Node,
                                 edge: Edge,
                                 node_features: to.Tensor) -> to.Tensor:
        edge_slice = edge.get_edge_slice()
        message_from_a_neighbor_other_than_target = messages[edge_slice]
        return to.sigmoid(
            to.add(
                to.add(self.w_gru_update_gate_features[edge_slice].matmul(node_features[node.node_id]),
                       self.u_gru_update_gate[edge_slice].matmul(message_from_a_neighbor_other_than_target)),
                self.b_gru_update_gate)).long()

    @staticmethod
    def _create_node(node_features: to.Tensor, adjacency_matrix: to.Tensor, node_id: to.Tensor) -> Node:
        return Node(node_features, adjacency_matrix, node_id)

    @staticmethod
    def _create_edge(start_node: Node, end_node: Node) -> Edge:
        return Edge(start_node, end_node)

    @staticmethod
    def _create_message(device: str) -> MessageGRU:
        return MessageGRU(device)



import torch as to


class Graph:
    def __init__(self, adjacency_matrix: to.Tensor, node_features: to.Tensor) -> None:
        self.adjacency_matrix = adjacency_matrix
        self.node_features = node_features
        self.number_of_nodes = self._get_number_of_nodes()
        self.number_of_node_features = self._get_number_of_node_features()

    def __eq__(self, o):
        return to.all(to.eq(self.adjacency_matrix, o.adjacency_matrix)) and \
               to.all(to.eq(self.node_features, o.node_features))

    def _get_number_of_nodes(self) -> int:
        return self.adjacency_matrix.shape[0]

    def _get_number_of_node_features(self) -> int:
        return self.node_features.shape[1]



import torch

optimizers = {
    "Adadelta": torch.optim.Adadelta,
    "Adagrad": torch.optim.Adagrad,
    "Adam": torch.optim.Adam,
    "AdamW": torch.optim.AdamW,
    "SparseAdam": torch.optim.SparseAdam,
    "Adamax": torch.optim.Adamax,
    "ASGD": torch.optim.ASGD,
    "LBFGS": torch.optim.LBFGS,
    "RMSprop": torch.optim.RMSprop,
    "Rprop": torch.optim.Rprop,
    "SGD": torch.optim.SGD,
}



import logging
import os
import pickle
from typing import List, Tuple

import torch as to

from message_passing_nn.repository.repository import Repository


class FileSystemRepository(Repository):
    def __init__(self, data_directory: str, dataset: str) -> None:
        super().__init__()
        self.data_directory = data_directory + dataset + '/'

    def save(self, filename: str, dataset_name: str) -> None:
        with open(self.data_directory + filename, 'wb') as file:
            pickle.dump(dataset_name, file)

    def get_all_features_and_labels_from_separate_files(self) -> List[Tuple[to.Tensor, to.Tensor]]:
        self.get_logger().info("Loading dataset")
        files_in_path = self._extract_name_prefixes_from_filenames()
        dataset = []
        for file in files_in_path:
            dataset.append((self._get_features(file), self._get_labels(file)))
        self.get_logger().info("Finished loading dataset")
        return dataset

    def _get_labels(self, filename: str) -> to.Tensor:
        with open(self.data_directory + filename + 'labels.pickle', 'rb') as labels_file:
            labels = pickle.load(labels_file).float()
        return labels

    def _get_features(self, filename: str) -> to.Tensor:
        with open(self.data_directory + filename + 'features.pickle', 'rb') as features_file:
            features = pickle.load(features_file).float()
        return features

    def _extract_name_prefixes_from_filenames(self):
        return set([self._reconstruct_filename(file) for file in self._get_data_filenames()])

    def _get_data_filenames(self):
        return sorted([file for file in os.listdir(self.data_directory) if file.endswith(".pickle")])

    @staticmethod
    def _reconstruct_filename(file):
        return "_".join(file.split("_")[:-1]) + "_"

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')



from unittest import TestCase

import numpy as np

from message_passing_nn.model.node import Node
from tests.fixtures.matrices_and_vectors import BASE_GRAPH_NODE_FEATURES, BASE_GRAPH


class TestNode(TestCase):
    def setUp(self) -> None:
        all_node_features = BASE_GRAPH_NODE_FEATURES
        labels = BASE_GRAPH.view(-1)
        self.node = Node(all_node_features, labels, node_id=2)

    def test_node_neighbors(self):
        # Given
        neighbors_expected = np.array([0, 1, 3])

        # When
        neighbors = self.node.neighbors

        # Then
        self.assertTrue(np.array_equal(neighbors_expected, neighbors))



import torch as to

from message_passing_nn.model.message import Message


class MessageGRU(Message):

    def __init__(self, device: str):
        super().__init__()
        self.previous_messages = None
        self.update_gate = None
        self.current_memory = None
        self.device = device

    def compose(self) -> None:
        self.value = to.add(
                            to.mul(
                                   to.sub(to.ones(self.update_gate.shape).to(self.device),
                                          self.update_gate),
                                   self.previous_messages),
                            to.mul(self.update_gate,
                                   self.current_memory))



from setuptools import setup, find_packages

setup(
    name='message-passing-nn',
    version='1.1.6',
    packages=find_packages(exclude=["tests"]),
    url='https://github.com/kovanostra/message-passing-nn',
    download_url='https://github.com/kovanostra/message-passing-nn/archive/1.1.6.tar.gz',
    keywords=['MESSAGE PASSING', 'NEURAL NETWORK', 'GRU'],
    license='MIT',
    author='Michail Kovanis',
    description='A message passing neural network with GRU units',
    install_requires=[
        'click',
        'pandas==1.0.3',
        'torch==1.4.0'
    ],
    entry_points={
        'console_scripts': [
            'message-passing-nn = message_passing_nn.cli:main'
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
)



import logging
import os
from datetime import datetime
from typing import Dict, List

import torch as to
from pandas import pandas as pd

from message_passing_nn.fixtures.filenames import *


class Saver:
    def __init__(self, model_directory: str, results_directory: str) -> None:
        self.model_directory = model_directory
        self.results_directory = results_directory

    def save_model(self, epoch: int, configuration_id: str, model: to.nn.Module) -> None:
        current_folder = self._join_path([self.model_directory, configuration_id])
        if not os.path.exists(current_folder):
            os.makedirs(current_folder)
        path_and_filename = self._join_path(
            [current_folder, self._join_strings([EPOCH, str(epoch), MODEL_STATE_DICTIONARY])])
        to.save(model.state_dict(), path_and_filename)
        self.get_logger().info("Saved model checkpoint in " + path_and_filename)

    def save_results(self, configuration_id: str, results: Dict) -> None:
        current_folder = self._join_path(
            [self.results_directory, configuration_id])
        if not os.path.exists(current_folder):
            os.makedirs(current_folder)
        results_dataframe = self._construct_dataframe_from_nested_dictionary(results)
        path_and_filename = self._join_path([current_folder,
                                             self._join_strings([datetime.now().strftime("%d-%b-%YT%H_%M"),
                                                                 RESULTS_CSV])])
        results_dataframe.to_csv(path_and_filename)
        self.get_logger().info("Saved results in " + path_and_filename)

    @staticmethod
    def _join_strings(fields: List) -> str:
        return "_".join(fields)

    @staticmethod
    def _construct_dataframe_from_nested_dictionary(results: Dict) -> pd.DataFrame:
        results_dataframe = pd.DataFrame.from_dict({(i, j): results[i][j]
                                                    for i in results.keys()
                                                    for j in results[i].keys()},
                                                   orient='index')
        return results_dataframe

    @staticmethod
    def _join_path(fields: List) -> str:
        return "/".join(fields)

    @staticmethod
    def get_logger() -> logging.Logger:
        return logging.getLogger('message_passing_nn')
